#!/usr/bin/env python3

"""
Generate config files for BlobToolKit pipeline.

Usage:
  btk pipeline generate-config <ACCESSION> [--coverage 30] [--download]
    [--out /path/to/output/directory] [--db /path/to/database/directory]
    [--db-suffix STRING] [--reads STRING...] [--read-runs INT] [--api-key STRING]
    [--platforms STRING]

Options:
  --coverage=INT         Maximum coverage for read mapping [default: 30]
  --download             Flag to download remote files [default: False]
  --out PATH             Path to output directory [default: .]
  --db PATH              Path to database directory [default: .]
  --db-suffix STRING     Database version suffix (e.g. 2021_06)
  --reads STRING         Read accession to include.
  --read-runs INT        Maximum number of read runs [default: 3]
  --api-key STRING       NCBI api key for use with edirect
  --platforms STRING     priority order for sequencing platforms
                         [default: PACBIO_SMRT,ILLUMINA_XTEN,ILLUMINA,OXFORD_NANOPORE,OTHER]
"""

import os
import re
import sys
from operator import itemgetter
from pathlib import Path
from subprocess import PIPE
from subprocess import Popen
from subprocess import run

import requests
import ujson
import yaml
from defusedxml import ElementTree as ET
from docopt import docopt
from tolkein import tofetch
from tolkein import tofile
from tolkein import tolog

LOGGER = tolog.logger(__name__)

GOAT_API = "https://goat.genomehubs.org/api/v0.0.1"
ENA_API = "https://www.ebi.ac.uk/ena/browser/api"
GCA_NAME = re.compile(r"GCA_.+")
BUSCO_URL = "https://busco-data.ezlab.org/v5/data/lineages"


def find_busco_lineages(ancestors):
    """Work out which BUSCO sets to run for a given taxid."""
    LOGGER.info("Identifying relevant BUSCO lineages")
    BUSCO_SETS = {
        "422676": "aconoidasida",
        "7898": "actinopterygii",
        "5338": "agaricales",
        "155619": "agaricomycetes",
        "33630": "alveolata",
        "5794": "apicomplexa",
        "6854": "arachnida",
        "6656": "arthropoda",
        "4890": "ascomycota",
        "8782": "aves",
        "5204": "basidiomycota",
        "68889": "boletales",
        "3699": "brassicales",
        "134362": "capnodiales",
        "33554": "carnivora",
        "91561": "cetartiodactyla",
        "34395": "chaetothyriales",
        "3041": "chlorophyta",
        "5796": "coccidia",
        "28738": "cyprinodontiformes",
        "7147": "diptera",
        "147541": "dothideomycetes",
        "3193": "embryophyta",
        "33392": "endopterygota",
        "314146": "euarchontoglires",
        "33682": "euglenozoa",
        "2759": "eukaryota",
        "5042": "eurotiales",
        "147545": "eurotiomycetes",
        "9347": "eutheria",
        "72025": "fabales",
        "4751": "fungi",
        "314147": "glires",
        "1028384": "glomerellales",
        "5178": "helotiales",
        "7524": "hemiptera",
        "7399": "hymenoptera",
        "5125": "hypocreales",
        "50557": "insecta",
        "314145": "laurasiatheria",
        "147548": "leotiomycetes",
        "7088": "lepidoptera",
        "4447": "liliopsida",
        "40674": "mammalia",
        "33208": "metazoa",
        "6029": "microsporidia",
        "6447": "mollusca",
        "4827": "mucorales",
        "1913637": "mucoromycota",
        "6231": "nematoda",
        "33183": "onygenales",
        "9126": "passeriformes",
        "5820": "plasmodium",
        "92860": "pleosporales",
        "38820": "poales",
        "5303": "polyporales",
        "9443": "primates",
        "4891": "saccharomycetes",
        "8457": "sauropsida",
        "4069": "solanales",
        "147550": "sordariomycetes",
        "33634": "stramenopiles",
        "32523": "tetrapoda",
        "155616": "tremellomycetes",
        "7742": "vertebrata",
        "33090": "viridiplantae",
        "71240": "eudicots",
    }
    # line = grep_line(lineage_file, taxid)
    lineages = []
    for obj in ancestors:
        if obj["taxon_id"] in BUSCO_SETS:
            lineages.append("%s_odb10" % BUSCO_SETS[obj["taxon_id"]])
    lineages.append("bacteria_odb10")
    lineages.append("archaea_odb10")
    return lineages


def fetch_file(url, filename):
    """fetch a remote file using aria2."""
    filepath = Path(filename)
    if filepath.is_file():
        LOGGER.info("File exists, not overwriting")
        return
    cmd = [
        "aria2c",
        "-q",
        "-d",
        filepath.parent,
        "-o",
        filepath.name,
        url,
    ]
    run(cmd)


def fetch_assembly_url(accession, api_key=None):
    """
    Fetch an assembly url using edirect.
    """
    eutils_env = os.environ.copy()
    if api_key and api_key is not None:
        eutils_env["NCBI_API_KEY"] = api_key
    esearch = Popen(
        "esearch -db assembly -query %s" % accession,
        stdout=PIPE,
        shell=True,
        env=eutils_env,
    )
    esummary = Popen(
        "esummary", stdin=esearch.stdout, stdout=PIPE, shell=True, env=eutils_env
    )
    xtract = Popen(
        "xtract -pattern DocumentSummary -element FtpPath_GenBank",
        stdin=esummary.stdout,
        stdout=PIPE,
        shell=True,
    )
    xtract_stdout = xtract.communicate()[0].decode("utf-8").strip()
    for url in xtract_stdout.split("\n"):
        basenames = re.findall(GCA_NAME, url)
        if basenames:
            for basename in basenames:
                if accession in basename:
                    return "%s/%s_genomic.fna.gz" % (url, basename)
    return None


def fetch_assembly_fasta(url, filename):
    """Save assembly fasta file to local disk."""
    LOGGER.info("Fetching assembly FASTA to %s" % filename)
    fetch_file(url, filename)


def parse_assembly_report(filename, cat_filename, syn_filename):
    """Parse synonyms and assembly level into tsv files."""
    synonyms = []
    categories = []
    cats = {
        "identifier": {"index": 4, "list": []},
        "assembly_role": {"index": 1, "list": []},
        "assembly_level": {"index": 3, "list": []},
        "assembly_unit": {"index": 7, "list": []},
    }
    names = {
        "identifier": {"index": 4, "list": []},
        "name": {"index": 0, "list": []},
        "assigned_name": {"index": 2, "list": []},
        "refseq_accession": {"index": 6, "list": []},
    }
    with tofile.open_file_handle(filename) as fh:
        for line in fh:
            if line.startswith("#"):
                continue
            row = line.rstrip().split("\t")
            for group in (cats, names):
                for obj in group.values():
                    value = row[obj["index"]]
                    obj["list"].append(value)
    header = []
    for key, obj in cats.items():
        if len(set(obj["list"])) > 1:
            header.append(key)
    categories.append(header)
    for idx, value in enumerate(cats[header[0]]["list"]):
        row = [value]
        for key in header[1:]:
            row.append(cats[key]["list"][idx])
        categories.append(row)
    tofile.write_file(cat_filename, categories)
    header = []
    for key, obj in names.items():
        if len(set(obj["list"])) > 1:
            header.append(key)
    synonyms.append(header)
    for idx, value in enumerate(names[header[0]]["list"]):
        row = [value]
        for key in header[1:]:
            row.append(names[key]["list"][idx])
        synonyms.append(row)
    tofile.write_file(syn_filename, synonyms)


def fetch_assembly_report(url, filename, cat_filename, syn_filename):
    """Save assembly report file to local disk."""
    LOGGER.info("Fetching assembly report to %s" % filename)
    fetch_file(url, filename)
    parse_assembly_report(filename, cat_filename, syn_filename)


def fetch_assembly_meta_xml(accession):
    """
    Fetch assembly metadata xml from ENA.
    """
    url = "%s/xml/%s" % (ENA_API, accession)
    xml = tofetch.fetch_url(url)
    return xml


def deep_find_text(data, tags):
    """
    Find nested attributes in xml.

    Return attribute value.
    """
    for tag in tags:
        try:
            data = data.find(tag)
        except:
            return ""
    return data.text


def parse_assembly_meta(accession):
    """Return dict of metadata values for an assembly."""
    LOGGER.info("Fetching assembly metadata")
    meta = {
        "assembly": {"accession": accession},
        "busco": {"lineages": []},
        "reads": {"paired": [], "single": []},
        "revision": 0,
        "settings": {
            "tmp": "/tmp",
            "blast_chunk": 100000,
            "blast_max_chunks": 10,
            "blast_overlap": 0,
            "blast_min_length": 1000,
            "stats_chunk": 1000,
            "stats_windows": [0.1, 0.01, 100000, 1000000],
        },
        "similarity": {
            "defaults": {
                "evalue": 1.0e-10,
                "import_evalue": 1.0e-25,
                "max_target_seqs": 10,
                "taxrule": "buscogenes",
            },
            "diamond_blastx": {"name": "reference_proteomes"},
            "diamond_blastp": {
                "name": "reference_proteomes",
                "import_max_target_seqs": 100000,
                "taxrule": "blastp=buscogenes",
            },
            "blastn": {"name": "nt"},
        },
        "taxon": {},
        "version": 1,
    }
    xml = fetch_assembly_meta_xml(accession)
    root = ET.fromstring(xml)
    asm = root.find("ASSEMBLY")
    meta["assembly"]["bioproject"] = deep_find_text(
        asm, ("STUDY_REF", "IDENTIFIERS", "PRIMARY_ID")
    )
    meta["assembly"]["biosample"] = deep_find_text(
        asm, ("SAMPLE_REF", "IDENTIFIERS", "PRIMARY_ID")
    )
    meta["taxon"]["taxid"] = deep_find_text(asm, ("TAXON", "TAXON_ID"))
    meta["taxon"]["name"] = deep_find_text(asm, ("TAXON", "SCIENTIFIC_NAME"))
    meta["assembly"]["level"] = asm.find("ASSEMBLY_LEVEL").text
    meta["assembly"]["alias"] = asm.attrib["alias"]
    wgs_prefix = deep_find_text(asm, ("WGS_SET", "PREFIX"))
    wgs_version = deep_find_text(asm, ("WGS_SET", "VERSION"))
    if wgs_prefix and wgs_version:
        meta["assembly"]["prefix"] = "%s%s" % (wgs_prefix, wgs_version.zfill(2))
    elif " " not in meta["assembly"]["alias"]:
        meta["assembly"]["prefix"] = meta["assembly"]["alias"].replace(".", "_")
    else:
        meta["assembly"]["prefix"] = meta["assembly"]["accession"].replace(".", "_")
    attributes = asm.find("ASSEMBLY_ATTRIBUTES")
    for attribute in attributes.findall("ASSEMBLY_ATTRIBUTE"):
        if attribute.find("TAG").text == "total-length":
            meta["assembly"]["span"] = int(attribute.find("VALUE").text)
        elif attribute.find("TAG").text == "scaffold-count":
            meta["assembly"]["scaffold-count"] = int(attribute.find("VALUE").text)
    return meta


def fetch_busco_lineages(busco_sets, buscodir):
    """Fetch busco lineages."""
    if not busco_sets:
        return
    lineages_to_fetch = []
    for lineage in busco_sets:
        busco_lineage = "%s/lineages/%s" % (buscodir, lineage)
        if not os.path.isdir(busco_lineage):
            lineages_to_fetch.append(lineage)
    if not lineages_to_fetch:
        return
    lineage_urls = {}
    LOGGER.info("Fetching BUSCO lineage directory listing")
    listing = tofetch.fetch_url("%s/" % BUSCO_URL)
    for entry in listing.split("\n"):
        parts = re.split(r"[\"\s]+", entry)
        if len(parts) == 8:
            busco_set = re.sub(r"\..+$", "", parts[2])
            lineage_urls.update({busco_set: "%s/%s" % (BUSCO_URL, parts[2])})
    for lineage in lineages_to_fetch:
        LOGGER.info("Fetching BUSCO lineage %s" % lineage)
        tofetch.fetch_tar(lineage_urls[lineage], buscodir)


def fetch_goat_data(taxon_id):
    """Fetch taxon metadata from GoaT."""
    LOGGER.info("Fetching taxon metadata")
    url = "%s/record?recordId=taxon_id-%s&result=taxon" % (GOAT_API, taxon_id)
    result = tofetch.fetch_url(url)
    if result is None:
        LOGGER.error("Unable to fetch taxon metadata for '%s' from GoaT", taxon_id)
        sys.exit(1)
    data = ujson.loads(result)
    return data["records"][0]["record"]


def fetch_read_info(accession, per_platform):
    """Fetch read info for an accession."""
    portal = "https://www.ebi.ac.uk/ena/portal/api"
    url = (
        "%s/filereport?accession=%s&result=read_run&fields=run_accession,fastq_bytes,base_count,library_strategy,library_selection,library_layout,instrument_platform,experiment_title,fastq_ftp"
        % (portal, accession)
    )
    data = tofetch.fetch_url(url)
    if data is None:
        return
    header = None
    for line in data.split("\n"):
        if not line or line == "":
            continue
        if header is None:
            header = line.split("\t")
            continue
        fields = line.split("\t")
        values = {}
        platform = "OTHER"
        for i in range(0, len(header)):
            value = fields[i]
            if header[i] == "instrument_platform" and platform != "ILLUMINA_XTEN":
                platform = fields[i]
            if header[i] == "experiment_title":
                if value == "HiSeq X Ten paired end sequencing":
                    platform = "ILLUMINA_XTEN"
            values.update({header[i]: value})
        if "base_count" in values:
            values["base_count"] = int(values["base_count"])
        else:
            values["base_count"] = 0
        try:
            per_platform[platform].append(values)
        except KeyError:
            per_platform["OTHER"].append(values)


def assembly_reads(accession, read_runs, platforms):
    """
    Query INSDC reads for an <accession> (or list of accessions).

    Return a dict of SRA accession, FASTQ ftp url, md5 and file size.
    """
    sra = []
    per_platform = {
        "PACBIO_SMRT": [],
        "ILLUMINA_XTEN": [],
        "ILLUMINA": [],
        "OXFORD_NANOPORE": [],
        "OTHER": [],
    }
    if not isinstance(accession, list):
        accession = [accession]
    for acc in accession:
        fetch_read_info(acc, per_platform)
    for key in platforms.split(","):
        arr = per_platform[key]
        runs = []
        if arr:
            for entry in arr:
                runs.append(entry)
        runs = sorted(runs, key=itemgetter("base_count"), reverse=True)[
            : read_runs - len(sra)
        ]
        sra += runs
        if len(sra) >= read_runs:
            break
    if sra:
        return sra
    return None


def base_count(x):
    """Return number of bases or zero."""
    if isinstance(x["base_count"], list):
        return int(x["base_count"][0] or 0)
    else:
        return 0


def fetch_read_files(meta):
    """Fetch sra reads."""
    files = meta["file"].split(";")
    for index, url in enumerate(meta["fastq_ftp"].split(";")):
        url = "ftp://%s" % url
        read_file = files[index]
        LOGGER.info("Fetching read file %s", read_file)
        fetch_file(url, read_file)


def add_taxon_to_meta(meta, taxon_meta):
    """Add taxon info to metadata."""
    LOGGER.info("Adding taxon metadata to assembly metadata")
    ranks = [
        "species",
        "genus",
        "family",
        "order",
        "class",
        "phylum",
        "kingdom",
        "superkingdom",
    ]
    for obj in taxon_meta["lineage"]:
        if obj["taxon_rank"] in ranks:
            meta["taxon"].update({obj["taxon_rank"]: obj["scientific_name"]})


def add_reads_to_meta(meta, sra, readdir):
    """Add read accessions to metadata."""
    LOGGER.info("Adding read accessions to assembly metadata")
    for index, library in enumerate(sra):
        strategy = library["library_layout"].lower()
        fastq_ftp = library["fastq_ftp"].split(";")[-2:]
        info = {
            "prefix": library["run_accession"],
            "platform": library["instrument_platform"],
            "base_count": library["base_count"],
            "file": ";".join(
                [re.sub(r"^.+/", "%s/" % readdir, url) for url in fastq_ftp]
            ),
            "url": ";".join(fastq_ftp),
        }
        library["file"] = info["file"]
        meta["reads"][strategy].append(info)


def set_btk_version(meta):
    """Get curent version of hosted datasets from BTK API."""
    LOGGER.info("Checking current version on BTK public viewer")
    string = meta["assembly"]["prefix"]
    btk = "https://blobtoolkit.genomehubs.org/api/v1/search/%s" % string
    response = requests.get(btk)
    current = 0
    if response.ok:
        data = yaml.full_load(response.text)
        for asm in data:
            if "version" in asm:
                current = asm["version"]
            else:
                current = 1
    meta["revision"] = current
    meta["version"] = current + 1


# def create_outdir(span, version=1, _lineage="all"):
#     """Create output directory."""
#     span = tolkein.tobin.readable_bin(span)
#     name = "%s/v%s/%s" % (OUTDIR, str(version), span)
#     os.makedirs("%s/" % name, exist_ok=True)
#     return name


def main():
    """Entry point."""
    opts = docopt(__doc__)
    accession = opts["<ACCESSION>"]
    outdir = opts["--out"]
    dbdir = opts["--db"]
    buscodir = "%s/busco" % dbdir
    uniprotdir = "%s/uniprot" % dbdir
    ntdir = "%s/nt" % dbdir
    taxdumpdir = "%s/taxdump" % dbdir
    if opts["--db-suffix"]:
        buscodir += "_%s" % opts["--db-suffix"]
        ntdir += "_%s" % opts["--db-suffix"]
        uniprotdir += "_%s" % opts["--db-suffix"]
        taxdumpdir += "_%s" % opts["--db-suffix"]
    if not outdir.endswith(accession):
        outdir += "/%s" % accession
    os.makedirs(outdir, exist_ok=True)
    meta = parse_assembly_meta(accession)
    assembly_url = fetch_assembly_url(accession, opts["--api-key"])
    if assembly_url is None:
        LOGGER.error("Unable to find assembly URL")
        sys.exit(1)
    assembly_file = "%s/assembly/%s.fasta.gz" % (outdir, accession)
    meta["assembly"].update({"file": assembly_file, "url": assembly_url})
    assembly_report = "%s/assembly/%s.report.txt" % (outdir, accession)
    syn_filename = "%s/assembly/%s.synonyms.tsv" % (outdir, accession)
    cat_filename = "%s/assembly/%s.categories.tsv" % (outdir, accession)
    meta["fields"] = {
        "synonyms": {"file": syn_filename, "prefix": "insdc"},
        "categories": {"file": cat_filename},
    }
    if opts["--download"]:
        os.makedirs(buscodir, exist_ok=True)
        os.makedirs("%s/assembly" % outdir, exist_ok=True)
        fetch_assembly_fasta(assembly_url, assembly_file)
        report_url = assembly_url.replace("_genomic.fna.gz", "_assembly_report.txt")
        fetch_assembly_report(report_url, assembly_report, cat_filename, syn_filename)
    taxon_meta = fetch_goat_data(meta["taxon"]["taxid"])
    add_taxon_to_meta(meta, taxon_meta)
    set_btk_version(meta)
    busco_sets = find_busco_lineages(taxon_meta["lineage"])
    if busco_sets:
        meta["busco"].update(
            {
                "download_dir": buscodir,
                "lineages": busco_sets,
                "basal_lineages": [
                    "eukaryota_odb10",
                    "bacteria_odb10",
                    "archaea_odb10",
                ],
            }
        )
    if opts["--download"]:
        fetch_busco_lineages(busco_sets, buscodir)
    read_accessions = []
    if meta["assembly"]["biosample"]:
        read_accessions = [meta["assembly"]["biosample"]]
    if opts["--reads"]:
        read_accessions += opts["--reads"]
    sra = assembly_reads(read_accessions, int(opts["--read-runs"]), opts["--platforms"])
    if sra:
        if opts["--coverage"]:
            meta["reads"].update({"coverage": {"max": int(opts["--coverage"])}})
        readdir = "%s/reads" % outdir
        add_reads_to_meta(meta, sra, readdir)
        if opts["--download"]:
            os.makedirs(readdir, exist_ok=True)
            for library in sra:
                fetch_read_files(library)
    meta["similarity"]["blastn"].update({"path": ntdir})
    meta["similarity"]["diamond_blastx"].update({"path": uniprotdir})
    meta["similarity"]["diamond_blastp"].update({"path": uniprotdir})
    meta["settings"]["taxdump"] = taxdumpdir
    tofile.write_file("%s/config.yaml" % outdir, meta)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# pylint: disable=no-member, too-many-branches, too-many-nested-blocks

"""
Add data to a BlobDir.

Usage:
    blobtools add [--bed BED...] [--beddir DIRECTORY] [--bedtsv TSV...] [--bedtsvdir DIRECTORY]
                  [--busco TSV...] [--cov BAM...] [--hits TSV...] [--fasta FASTA] [--hits-cols LIST]
                  [--key path=value...] [--link path=url...] [--taxid INT] [--skip-link-test]
                  [--blobdb JSON] [--meta YAML] [--synonyms TSV...] [--trnascan TSV...]
                  [--text TXT...] [--text-delimiter STRING] [--text-cols LIST] [--text-header]
                  [--text-no-array] [--taxdump DIRECTORY] [--taxrule bestsum|bestsumorder[=prefix]]
                  [--threads INT] [--evalue NUMBER] [--bitscore NUMBER] [--hit-count INT]
                  [--update-plot] [--pileup-args key=value...] [--create] [--replace] DIRECTORY

Arguments:
    DIRECTORY             Existing Blob directory.

Options:
    --bed BED             BED format file.
    --beddir DIRECTORY    Directory containing one or more BED format files.
    --bedtsv TSV          TSV file with header row and bed-format columns 1-3.
    --bedtsvdir DIRECTORY Directory containing one or more BED-like tsv files.
    --busco TSV           BUSCO full_table.tsv output file.
    --cov BAM             BAM/SAM/CRAM read alignment file.
    --fasta FASTA         FASTA sequence file.
    --hits TSV            Tabular BLAST/Diamond output file.
    --hits-cols LIST      Comma separated list of <column number>=<field name>.
                          [Default: 1=qseqid,2=staxids,3=bitscore,5=sseqid,10=qstart,11=qend,14=evalue]
    --taxid INT           Add ranks to metadata for a taxid.
    --key path=value      Set a metadata key to value.
    --link path=URL       Link to an external resource.
    --skip-link-test      Skip test to see if link URL can be resolved.
    --meta YAML           Dataset metadata.
    --blobdb JSON         Blobtools v1 blobDB.
    --synonyms TSV        TSV file containing current identifiers and synonyms.
    --taxdump DIRECTORY   Location of NCBI new_taxdump directory.
    --taxrule rulename[=prefix]
                          Rule to use when assigning BLAST hits to taxa (bestsum, bestsumorder,
                          bestdistsum, bestdistsumorder, blastp).
                          An alternate prefix may be specified. [Default: bestsumorder]
    --threads INT         Number of threads to use for multithreaded tasks. [Default: 1]
    --evalue FLOAT        Set evalue cutoff when parsing hits file. [Default: 1]
    --bitscore FLOAT      Set bitscore cutoff when parsing hits file. [Default: 1]
    --hit-count INT       Number of hits to parse when inferring taxonomy. [Default: 10]
    --update-plot         Flag to use new taxrule as default category.
    --text TXT            Generic text file.
    --text-delimiter STRING
                          Text file delimiter. [Default: whitespace]
    --text-cols LIST      Comma separated list of <column number>[=<field name>].
    --text-header         Flag to indicate first row of text file contains field names.
    --text-no-array       Flag to prevent fields in files with duplicate identifiers being
                          loaded as array fields.
    --trnascan TSV        tRNAscan2-SE output
    --pileup-args key=val Key/value pairs to pass to samtools pileup.
    --create              Create a new BlobDir.
    --replace             Replace existing fields with matching ids.

Examples:
    # 1. Add BUSCO scores to BlobDir
    ./blobtools add --busco busco.full_table.tsv BlobDir

"""

import sys

from docopt import docopt

from ..lib import bed
from ..lib import blob_db
from ..lib import busco
from ..lib import cov
from ..lib import fasta
from ..lib import file_io
from ..lib import hits
from ..lib import key
from ..lib import link
from ..lib import synonyms
from ..lib import taxid
from ..lib import text
from ..lib import trnascan
from .fetch import fetch_field
from .fetch import fetch_metadata
from .fetch import fetch_taxdump
from .field import Identifier
from .version import __version__

FIELDS = [
    {"flag": "--bed", "module": bed, "optional": ["identifiers"]},
    {"flag": "--beddir", "module": bed, "optional": ["identifiers"]},
    {"flag": "--bedtsv", "module": bed, "optional": ["identifiers"]},
    {"flag": "--bedtsvdir", "module": bed, "optional": ["identifiers"]},
    {"flag": "--fasta", "module": fasta, "optional": ["identifiers"]},
    {"flag": "--blobdb", "module": blob_db, "depends": ["identifiers"]},
    {"flag": "--busco", "module": busco, "depends": ["identifiers"]},
    {"flag": "--text", "module": text, "depends": ["identifiers"]},
    {"flag": "--trnascan", "module": trnascan, "depends": ["identifiers"]},
    {"flag": "--cov", "module": cov, "depends": ["identifiers", "length", "ncount"]},
    {"flag": "--hits", "module": hits, "depends": ["identifiers", "length"]},
    {"flag": "--synonyms", "module": synonyms, "depends": ["identifiers"]},
]
PARAMS = set(
    ["--taxrule", "--threads", "--pileup-args", "--evalue", "--bitscore", "--hit-count"]
)


def has_field_warning(meta, field_id):
    """Warn if dataset has existing field with same id."""
    if meta.has_field(field_id):
        print(
            "WARN: Field '%s' is already present in dataset, not overwriting."
            % field_id
        )
        print("WARN: Use '--replace' flag to overwrite existing field.")
        return 1
    return 0


def main(args):
    """Entrypoint for blobtools add."""
    meta = fetch_metadata(args["DIRECTORY"], **args)
    if args["--fasta"]:
        meta.assembly.update({"file": args["--fasta"]})
    taxdump = None
    dependencies = {}
    for field in FIELDS:
        if args[field["flag"]]:
            if "depends" in field:
                for dep in field["depends"]:
                    if dep not in dependencies or not dependencies[dep]:
                        dependencies[dep] = fetch_field(args["DIRECTORY"], dep, meta)
            for dep_key, dep_value in dependencies.items():
                if not dep_value:
                    print("ERROR: '%s.json' was not found in the BlobDir." % dep_key)
                    print(
                        "ERROR: You may need to rebuild the BlobDir to run this command."
                    )
                    sys.exit(1)
            if field["flag"] == "--hits":
                if not taxdump:
                    taxdump = fetch_taxdump(args["--taxdump"])
            parents = field["module"].parent()
            if "optional" in field:
                for dep in field["optional"]:
                    if dep not in dependencies or not dependencies[dep]:
                        dependencies[dep] = fetch_field(args["DIRECTORY"], dep, meta)
            parsed = field["module"].parse(
                args[field["flag"]],
                **{key: args[key] for key in args.keys()},
                taxdump=taxdump,
                dependencies=dependencies,
                meta=meta
            )
            if not isinstance(parsed, list):
                parsed = [parsed]
            for data in parsed:
                if not args["--replace"]:
                    if has_field_warning(meta, data.field_id):
                        continue
                for parent in data.parents:
                    if "range" in parent:
                        parent_meta = meta.field_meta(parent["id"])
                        if parent_meta and "range" in parent_meta:
                            parent["range"][0] = min(
                                parent["range"][0], parent_meta["range"][0]
                            )
                            parent["range"][1] = max(
                                parent["range"][1], parent_meta["range"][1]
                            )
                meta.add_field(parents + data.parents, **data.meta)
                if isinstance(data, Identifier):
                    meta.records = len(data.values)
                json_file = "%s/%s.json" % (args["DIRECTORY"], data.field_id)
                file_io.write_file(json_file, data.values_to_dict())
                dependencies[data.field_id] = data
    if "identifiers" not in dependencies:
        dependencies["identifiers"] = fetch_field(
            args["DIRECTORY"], "identifiers", meta
        )
    for string in args["--link"]:
        link.add(
            string, meta, dependencies["identifiers"].values, args["--skip-link-test"]
        )
    for string in args["--key"]:
        key.add(string, meta, args["--replace"])
    if args["--taxid"]:
        if not taxdump:
            taxdump = fetch_taxdump(args["--taxdump"])
        taxid.add(args["--taxid"], taxdump, meta)
    file_io.write_file("%s/meta.json" % args["DIRECTORY"], meta.to_dict())


def cli():
    """Entry point."""
    name = __name__.split(".")[-1]
    command_index = sys.argv.index(name)
    if len(sys.argv) == command_index + 1:
        args = docopt(__doc__, argv=[])
    else:
        args = docopt(__doc__, version=__version__)
    main(args)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3

# pylint: disable=c-extension-no-member

"""
Read, write and parse files.

Simple tests are included in docstrings and run using doctest.

Further tests are implemented in a separate file using pytest
(tests/unit_tests/test_file_io.py) to avoid adding test-specific code to
docstrings.
"""

import csv
import gzip
import io
import os
import pathlib
import sys
from itertools import groupby
from subprocess import PIPE
from subprocess import Popen

import ujson
import yaml


def read_file(filename):
    r"""
    Read a whole file into memory.

    >>> read_file('tests/files/infile')
    'testfile content\n'
    >>> read_file('nofile')

    """
    try:
        with stream_file(filename) as fh:
            return fh.read()
    except AttributeError:
        return None


def stream_file(filename):
    """
    Stream a file, line by line.

    Automatically detect gzipped files based on suffix.
    """
    if ".gz" in pathlib.Path(filename).suffixes:
        try:
            return gzip.open(filename, "rt")
        except OSError:
            return None
    try:
        return open(filename, "r")
    except IOError:
        return None


def stream_fasta(filename):
    """
    Stream a fasta file, sequence by sequence.

    Automatically detect gzipped files based on suffix.
    """
    if ".gz" in pathlib.Path(filename).suffixes:
        cmd = ["pigz", "-dc", filename]
    else:
        cmd = ["cat", filename]
    with Popen(cmd, stdout=PIPE, encoding="utf-8", bufsize=4096) as proc:
        faiter = (x[1] for x in groupby(proc.stdout, lambda line: line[0] == ">"))
        for header in faiter:
            seq_id = header.__next__().split()[0].replace(">", "")
            seq_str = "".join(map(lambda s: s.strip(), faiter.__next__()))
            yield seq_id, seq_str


def load_yaml(filename):
    """
    Parse a JSON/YAML file.

    load_yaml('identifiers.yaml')
    """
    data = read_file(filename)
    if data is None:
        return data
    if ".json" in filename:
        content = ujson.loads(data)
    elif ".yaml" in filename:
        content = yaml.full_load(data)
    else:
        content = data
    return content


def write_file(filename, data, plain=False):  # pylint: disable=too-many-branches
    """
    Write a file, use suffix to determine type and compression.

    - types: '.json', '.yaml'
    - compression: None, '.gz'

    write_file('variable.json.gz')
    """
    if ".json" in filename:
        content = ujson.dumps(data, indent=1, escape_forward_slashes=False)
    elif ".yaml" in filename:
        content = yaml.dump(data, indent=1)
    elif filename == "STDOUT":
        sys.stdout.write(
            ujson.dumps(data, indent=1, escape_forward_slashes=False) + "\n"
        )
        return True
    elif filename == "STDOUT":
        sys.stderr.write(
            ujson.dumps(data, indent=1, escape_forward_slashes=False) + "\n"
        )
        return True
    elif plain:
        content = "\n".join(data)
    elif ".csv" in filename or ".tsv" in filename:
        output = io.StringIO()
        if ".csv" in filename:
            writer = csv.writer(output, quoting=csv.QUOTE_NONNUMERIC)
        else:
            writer = csv.writer(output, delimiter="\t")
        for row in data:
            writer.writerow(row)
        content = output.getvalue()
    else:
        content = data
    if ".gz" in filename:
        try:
            with gzip.open(filename, "wt") as fh:
                fh.write(content)
        except OSError:
            return False
    else:
        try:
            with open(filename, "wt") as fh:
                fh.write(content)
        except IOError:
            return False
    return True


def delete_file(file):
    """Delete a file if exists."""
    if os.path.exists(file):
        os.remove(file)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#!/usr/bin/env python3
"""Parse BUSCO results into MultiArray Field."""

import re
from collections import defaultdict

from ..lib import file_io
from .field import MultiArray


def parse_trnascan(trnascan_file, identifiers):
    """Parse tRNAscan results into a MultiArray."""
    data = file_io.read_file(trnascan_file)
    lines = data.split("\n")
    header = True
    meta = {"file": trnascan_file}
    results = defaultdict(list)
    for line in lines:
        if header:
            row = re.split(" +", line)
            if len(row) > 1:
                if row[1].startswith("v."):
                    meta.update({"version": row[1]})
                elif row[1] == "Mode:":
                    meta.update({"mode": row[2]})
                    meta.update({"field_id": "trnascan_%s" % row[2].lower()})
                elif row[1].startswith("------"):
                    header = False
        else:
            row = re.split(r" +|\t", line)
            if len(row) == 9:
                results[row[0]].append([row[4], row[5]])
    if not identifiers.validate_list(list(results.keys())):
        raise UserWarning(
            "Contig names in the tRNAScan file did not match dataset identifiers."
        )
    values = [results[id] if id in results else [] for id in identifiers.values]
    trnascan_field = MultiArray(
        meta["field_id"],
        values=values,
        meta=meta,
        headers=("tRNA_type", "Anticodon"),
        parents=["children"],
    )
    return trnascan_field


def parse(files, **kwargs):
    """Parse all tRNAscan files."""
    parsed = []
    for file in files:
        parsed.append(
            parse_trnascan(file, identifiers=kwargs["dependencies"]["identifiers"])
        )
    return parsed


def parent():
    """Set standard metadata for tRNAscan."""
    trnascan = {
        "datatype": "mixed",
        "type": "array",
        "id": "trnascan",
        "name": "tRNAscan",
    }
    return [trnascan]



#!/usr/bin/env python3
"""Taxdump Class module."""

import os
import re

from ..lib import file_io


class Taxdump:
    """Class for working with NCBI taxonomy."""

    __slots__ = ["directory", "ancestors", "names", "ranks"]

    def __init__(self, directory, **kwargs):
        """Init Taxdump class."""
        self.directory = directory
        self.ancestors = {}
        self.ranks = {}
        self.names = {}
        if kwargs:
            self.update_data(**kwargs)
        else:
            self.load_ranks()
            self.load_names()
            self.load_ancestors()

    def update_data(self, **kwargs):
        """Update values and keys for an existing field."""
        for key, value in kwargs.items():
            setattr(self, key, {int(taxid): data for taxid, data in value.items()})

    def load_ranks(self):
        """Load ranks from file."""
        filename = os.path.abspath(os.path.join(self.directory, "nodes.dmp"))
        try:
            for line in file_io.stream_file(filename):
                row = self.parse_taxdump_row(line)
                if len(row) > 1:
                    self.ranks[int(row[0])] = row[2]
        except TypeError:
            print("ERROR: Unable to parse %s." % filename)
            exit(1)

    def load_names(self):
        """Load names from file."""
        filename = os.path.abspath(os.path.join(self.directory, "names.dmp"))
        for line in file_io.stream_file(filename):
            row = self.parse_taxdump_row(line)
            if row and row[3] == "scientific name":
                self.names[int(row[0])] = row[1]

    def load_ancestors(self):
        """Load ancestors from file."""
        filename = os.path.abspath(os.path.join(self.directory, "taxidlineage.dmp"))
        for line in file_io.stream_file(filename):
            row = self.parse_taxdump_row(line)
            if row[1]:
                taxid = int(row[0])
                self.ancestors[taxid] = {
                    self.ranks[int(id)]: int(id)
                    for id in row[1].split(" ")
                    if self.ranks[int(id)] in self.list_ranks()
                }
                if self.ranks[taxid] in self.list_ranks():
                    self.ancestors[taxid].update({self.ranks[taxid]: taxid})
                last = 0
                for rank in self.list_ranks():
                    if rank in self.ancestors[taxid]:
                        last = -self.ancestors[taxid][rank]
                    else:
                        self.ancestors[taxid].update({rank: last})

    def values_to_dict(self):
        """Create a dict of values."""
        data = {}
        for key in ("ancestors", "names", "ranks"):
            if hasattr(self, key):
                data[key] = getattr(self, key)
        return data

    def lineage(self, taxid):
        """Create a dict of rank-name pairs for a taxid."""
        lineages = {}
        try:
            ancestors = self.ancestors[taxid]
        except ValueError:
            return {}
        for rank in self.list_ranks():
            if rank in ancestors and ancestors[rank] > 0:
                lineages.update({rank: self.names[ancestors[rank]]})
        return lineages

    @staticmethod
    def parse_taxdump_row(line):
        """Parse an ncbi taxdump file."""
        return re.split(r"\s*\|\s*", line)[:-1]

    @staticmethod
    def list_ranks():
        """Return a list of taxonomic ranks."""
        return [
            "superkingdom",
            "kingdom",
            "phylum",
            "class",
            "order",
            "family",
            "genus",
            "species",
        ]


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#!/usr/bin/env python3

"""
Window stats.

Usage:
    btk pipeline window-stats --in TSV [--window FLOAT...] [--min-window-length INT] [--min-window-count INT] --out TSV

Options:
    --in TSV                 chunked summary stats tsv file.
    --min-window-length INT  minimum length of a window. [Default: 100000]
    --min-window-count INT   minimum number of windows. [Default: 1]
    --window FLOAT           window size or proportion. [Default: 1]
    --out TSV                output TSV filename or suffix.
"""

import logging
import math
import re
import statistics
from collections import defaultdict
from pathlib import Path

from docopt import DocoptExit
from docopt import docopt
from tolkein import tofile

logger_config = {
    "level": logging.INFO,
    "format": "%(asctime)s [%(levelname)s] line %(lineno)d %(message)s",
    "filemode": "w",
}
logging.basicConfig(**logger_config)
logger = logging.getLogger()


# def parse_args():
#     """Parse snakemake args if available."""
#     args = {}
#     try:
#         args["--in"] = snakemake.input.tsv
#         args["--window"] = str(snakemake.params.window)
#         args["--out"] = snakemake.output.tsv
#         for key, value in args.items:
#             sys.argv.append(key)
#             sys.argv.append(value)
#     except NameError as err:
#         pass


def parse_chunked_values(filename):
    """Parse chunked values into dict."""
    interval = 0
    header = None
    values = defaultdict(lambda: defaultdict(list))
    lengths = defaultdict(int)
    with tofile.open_file_handle(filename) as fh:
        for line in fh.readlines():
            row = line.rstrip().split("\t")
            if header is None:
                header = {key: idx + 3 for idx, key in enumerate(row[3:])}
                continue
            seqid = row[0]
            chunk_length = int(row[2]) - int(row[1])
            if chunk_length > interval:
                interval = chunk_length
            lengths[seqid] += chunk_length
            for key, idx in header.items():
                values[seqid][key].append(float(row[idx]))
    return lengths, values, interval


def round_to_interval(value, interval):
    """Round a number to a given interval."""
    return round(
        round(value / interval + 0.5) * interval, -int(math.floor(math.log10(interval)))
    )


def get_window_size(length, interval, window, min_length, min_count):
    """Get size of window based on options."""
    if window == 1:
        return length
    if window < 1:
        window_length = round_to_interval(length * window, interval)
        if window_length > min_length:
            return window_length
        return None
    if length / window >= min_count:
        return round_to_interval(window, interval)
    return None


def calculate_mean(arr, log):
    """Calculate mean and sd of arr values."""
    n = len(arr)
    if log:
        logged_arr = [math.log10(value) for value in arr if value > 0]
        if not logged_arr:
            return 0, 0, n
        mean = math.pow(10, statistics.mean(logged_arr))
        if len(logged_arr) > 1:
            sd = math.pow(10, statistics.stdev(logged_arr))
        else:
            sd = 0
        return mean, sd, n
    mean = statistics.mean(arr)
    if n > 1:
        sd = statistics.stdev(arr)
    else:
        sd = 0
    return mean, sd, n


def calculate_window_stats(lengths, chunks, window, interval, args):
    """Calculate mean and sd in windows across each sequence."""
    values = defaultdict(dict)
    for seqid, length in lengths.items():
        window_size = get_window_size(
            length,
            interval,
            window,
            int(args["--min-window-length"]),
            int(args["--min-window-count"]),
        )
        if window_size is not None:
            for key, arr in chunks[seqid].items():
                start_i = 0
                start_pos = 0
                while start_pos < length:
                    end_pos = min(start_pos + window_size, length)
                    mid_pos = start_pos + (end_pos - start_pos) / 2
                    end_i = round(end_pos / interval + 0.5) - 1
                    if window == 1:
                        proportion = "1"
                    else:
                        proportion = "%.3f" % (mid_pos / length)
                    if start_pos not in values[seqid]:
                        values[seqid][start_pos] = {
                            "end": str(end_pos),
                            "proportion": proportion,
                        }
                    if key.endswith("count"):
                        values[seqid][start_pos][key] = "%d" % sum(arr[start_i:end_i])
                    else:
                        mean, sd, n = calculate_mean(
                            arr[start_i : end_i + 1], key.endswith("_cov")
                        )
                        values[seqid][start_pos][key] = "%.3f" % mean
                        values[seqid][start_pos]["%s_sd" % key] = "%.3f" % sd
                        values[seqid][start_pos]["%s_n" % key] = "%d" % n
                    start_pos = end_pos
                    start_i = end_i + 1
    return values


def main():
    """Entry point."""
    try:
        # parse_args()
        args = docopt(__doc__)
    except DocoptExit:
        raise DocoptExit
    try:
        lengths, chunks, interval = parse_chunked_values(args["--in"])
        outfile = args["--out"]
        if outfile.startswith("."):
            outfile = "%s%s" % (args["--in"], args["--out"])
        outfile = Path(outfile)
        suffix = outfile.suffix
        filename = outfile.stem
        for window in args["--window"]:
            window = float(window)
            values = calculate_window_stats(lengths, chunks, window, interval, args)
            rows = []
            header = None
            for seqid, obj in values.items():
                for start_pos, entry in obj.items():
                    if header is None:
                        header = ["sequence", "start", "end"]
                        for key in entry.keys():
                            if key not in header:
                                header.append(key)
                        rows.append("\t".join(header) + "\n")
                    row = [seqid, str(start_pos)]
                    for key in header[2:]:
                        row.append(entry[key])
                    rows.append("\t".join(row) + "\n")
            filetag = ""
            if window != 1:
                filetag = ".%s" % re.sub(r"\.0$", "", str(window))
            with open("%s%s%s" % (filename, filetag, suffix), "w") as fh:
                fh.writelines(rows)
    except Exception as err:
        logger.error(err)
        exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# pylint: disable=no-member, too-many-branches, too-many-statements, too-many-locals, W0603, W0703

"""
Host a collection of BlobDirs.

Usage:
    blobtools host [--port INT]  [--api-port INT]
                   [--hostname STRING] DIRECTORY

Arguments:
    DIRECTORY             Directory containing one or more BlobDirs.

Options:
    --port INT            HTTP port number. [Default: 8080]
    --api-port INT        API port number. [Default: 8000]
    --hostname STRING     Hostname used to connect to API. [Default: localhost]
"""

import os
import platform
import shlex
import signal
import socket
import sys
import time
from pathlib import Path
from shutil import which
from subprocess import PIPE
from subprocess import Popen

import psutil
from docopt import docopt

from .version import __version__

PIDS = []


def iter_user_procs(process):
    """Iterate through processes owned by the current user."""
    username = psutil.Process.username(
        process
    )  # get username in format used by processes
    for proc in psutil.process_iter(["username"]):
        if proc.info["username"] == username:
            yield proc


def close_ports(process, args):
    """Close processes running on chosen ports."""
    for proc in iter_user_procs(process):
        try:
            for conns in proc.connections(kind="inet"):
                if conns.laddr.port == args["--api-port"]:
                    proc.send_signal(signal.SIGTERM)
                elif conns.laddr.port == args["--port"]:
                    proc.send_signal(signal.SIGTERM)
        except psutil.NoSuchProcess:
            continue
        except psutil.ZombieProcess:
            continue
        except psutil.AccessDenied:
            continue


def kill_child_processes(parent_pid, args=None, sig=signal.SIGTERM):
    """Kill all child processes."""
    try:
        parent = psutil.Process(parent_pid)
    except psutil.NoSuchProcess:
        return
    # if args:
    #     close_ports(parent, args)
    children = parent.children(recursive=True)
    for process in children:
        try:
            process.send_signal(sig)
        except psutil.NoSuchProcess:
            continue
        except psutil.ZombieProcess:
            continue
    parent.send_signal(sig)


def test_port(port, service):
    """Exit if port is already in use."""
    port = int(port)
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as skt:
        try:
            skt.bind(("", port))
        except OSError:
            if service == "test":
                return False
            print("ERROR: Port %d already in use, unable to host %s." % (port, service))
            print(
                "       Use: `lsof -nP -iTCP:%d | grep LISTEN` to find the associated process."
                % port
            )
            print(
                "       It may take ~30s for this port to become available when restarting %s."
                % service
            )
            sys.exit(1)
    return True


def find_binary(tool):
    """Find a binary executable for the blobtoolkit viewer or API."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    system = platform.system()
    # arch, _ = platform.architecture()
    default_binaries = {"api": "blobtoolkit-api", "viewer": "blobtoolkit-viewer"}
    if system == "Linux":
        binaries = {
            "api": "blobtoolkit-api-linux",
            "viewer": "blobtoolkit-viewer-linux",
        }
    if system == "Windows":
        binaries = {
            "api": "blobtoolkit-api-win.exe",
            "viewer": "blobtoolkit-viewer-win.exe",
        }
        default_binaries = {
            "api": "blobtoolkit-api.exe",
            "viewer": "blobtoolkit-viewer.exe",
        }
    if system == "Darwin":
        binaries = {
            "api": "blobtoolkit-api-macos",
            "viewer": "blobtoolkit-viewer-macos",
        }
    if which(default_binaries[tool]) is not None:
        return default_binaries[tool]
    if which(binaries[tool]) is not None:
        return default_binaries[tool]
    executable = binaries[tool]
    executable_path = os.path.join(script_dir, "bin", executable)
    if os.path.isfile(executable_path):
        return executable_path
    else:
        print(
            "ERROR: %s executable was not found. Please add %s to your PATH."
            % (default_binaries[tool], default_binaries[tool])
        )
        sys.exit(1)


def start_api(port, api_port, hostname, directory):
    """Start BlobToolKit API."""
    cmd = find_binary("api")
    # cmd = "blobtoolkit-api"
    origins = "http://localhost:%d http://localhost null" % int(port)
    if hostname != "localhost":
        origins += " http://%s:%d http://%s" % (hostname, int(port), hostname)
    if directory == "_":
        env = dict(os.environ, BTK_API_PORT=api_port, BTK_ORIGINS=origins,)
    else:
        env = dict(
            os.environ,
            BTK_API_PORT=api_port,
            BTK_FILE_PATH=directory,
            BTK_ORIGINS=origins,
        )
    process = Popen(
        shlex.split(cmd), stdout=PIPE, stderr=PIPE, encoding="ascii", env=env,
    )
    return process


def start_viewer(port, api_port, hostname):
    """Start BlobToolKit viewer."""
    cmd = find_binary("viewer")
    # cmd = "blobtoolkit-viewer"
    api_url = "http://%s:%d/api/v1" % (hostname, int(api_port))
    process = Popen(
        shlex.split(cmd),
        stdout=PIPE,
        stderr=PIPE,
        encoding="ascii",
        env=dict(
            os.environ,
            BTK_HOST=hostname,
            BTK_CLIENT_PORT=port,
            BTK_API_PORT=api_port,
            BTK_API_URL=api_url,
        ),
    )
    return process


def main(args):
    """Entrypoint for blobtools host."""
    global PIDS
    directory = args["DIRECTORY"]
    if directory != "_":
        path = Path(directory)
        if not path.exists():
            print("ERROR: Directory '%s' does not exist" % directory)
            sys.exit(1)
        if (path / "meta.json").exists():
            print("WARNING: Directory '%s' appears to be a BlobDir." % directory)
            print("         Hosting the parent directory instead.")
            path = path.resolve().parent
        directory = path.absolute()
    test_port(args["--api-port"], "BlobtoolKit API")
    test_port(args["--port"], "BlobtoolKit viewer")
    api = start_api(args["--port"], args["--api-port"], args["--hostname"], directory,)
    PIDS.append(api.pid)
    print(
        "Starting BlobToolKit API on port %d (pid: %d)"
        % (int(args["--api-port"]), api.pid)
    )
    time.sleep(2)
    viewer = start_viewer(args["--port"], args["--api-port"], args["--hostname"])
    PIDS.append(viewer.pid)
    print(
        "Starting BlobToolKit viewer on port %d (pid: %d)"
        % (int(args["--port"]), viewer.pid)
    )
    time.sleep(2)
    ready = False
    url = "http://%s:%d" % (args["--hostname"], int(args["--port"]))
    while True:
        time.sleep(1)
        if api.poll() is not None:
            for line in api.stdout.readlines():
                print(line.strip())
            for line in api.stderr.readlines():
                print(line.strip())
            if viewer.poll() is not None:
                for line in viewer.stdout.readlines():
                    print(line.strip())
                for line in viewer.stderr.readlines():
                    print(line.strip())
                try:
                    os.kill(viewer.pid, signal.SIGTERM)
                except ProcessLookupError:
                    pass
            break
        if viewer.poll() is not None:
            for line in viewer.stdout.readlines():
                print(line.strip())
            for line in viewer.stderr.readlines():
                print(line.strip())
            try:
                os.kill(api.pid, signal.SIGTERM)
            except ProcessLookupError:
                pass
            break
        if not ready:
            print("Visit %s to use the interactive BlobToolKit Viewer." % url)
            ready = True
        time.sleep(1)


def cli():
    """Entry point."""
    if len(sys.argv) == sys.argv.index(__name__.split(".")[-1]) + 1:
        args = docopt(__doc__, argv=[])
    else:
        args = docopt(__doc__, version=__version__)
    try:
        main(args)
    except KeyboardInterrupt:
        pass
    finally:
        for pid in PIDS:
            kill_child_processes(pid, args=args, sig=signal.SIGKILL)
            time.sleep(2)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
"""Functions to fetch data and metadata."""

import glob
import os
from pathlib import Path

from ..lib import file_io
from .dataset import Metadata
from .field import Array
from .field import Category
from .field import Field
from .field import Identifier
from .field import MultiArray
from .field import Variable
from .taxdump import Taxdump

TYPES = {
    "identifier": Identifier,
    "category": Category,
    "variable": Variable,
    "array": Array,
    "multiarray": MultiArray,
    "field": Field,
}


def fetch_field(path_to_dataset, field_id, meta=None):
    """
    Load fields from file.

    fetch_field('tests/files/dataset', 'identifiers', meta)
    """
    field_meta = meta.field_meta(field_id)
    try:
        data = file_io.load_yaml("%s/%s.json" % (path_to_dataset, field_id))
        if data is None:
            data = file_io.load_yaml("%s/%s.json.gz" % (path_to_dataset, field_id))
        if data is not None:
            data.update({"meta": field_meta})
        field = TYPES[field_meta["type"]](field_id, **data)
    except TypeError:
        field = False
    except KeyError:
        field = False
    return field


def fetch_metadata(path_to_dataset, **kwargs):
    """
    Load Metadata from file.

    fetch_metadata('tests/files/dataset')
    """
    dataset_id = path_to_dataset.split("/").pop()
    new_meta = {}
    meta = None
    if not os.path.exists(path_to_dataset):
        os.makedirs(path_to_dataset)
    if kwargs.get("--meta"):
        new_meta = file_io.load_yaml(kwargs["--meta"])
        if (kwargs["--bed"] or kwargs["--fasta"]) and kwargs["--replace"]:
            files = glob.glob("%s/*" % kwargs["DIRECTORY"])
            for file in files:
                os.remove(file)
    try:
        meta = kwargs["meta"]
    except KeyError:
        try:
            meta = file_io.load_yaml("%s/meta.json" % path_to_dataset)
            if meta is None:
                meta = file_io.load_yaml("%s/meta.json.gz" % path_to_dataset)
        except ValueError:
            pass
    if meta is None:
        meta = {}
    if "id" not in meta:
        meta["id"] = dataset_id
        meta["name"] = dataset_id
    for key, value in new_meta.items():
        if isinstance(value, dict):
            try:
                meta[key].update({k: v for k, v in value.items()})
            except KeyError:
                meta[key] = value
        elif isinstance(value, list):
            meta[key] += value
        else:
            meta[key] = value
    return Metadata(dataset_id, **meta)


def fetch_taxdump(path_to_taxdump):
    """Load Taxdump from file."""
    json_file = "%s/taxdump.json" % path_to_taxdump
    if not Path(json_file).exists():
        print("Parsing taxdump")
    else:
        print("Loading parsed taxdump")
    data = file_io.load_yaml(json_file)
    if data is None:
        taxdump = Taxdump(path_to_taxdump)
        file_io.write_file(json_file, taxdump.values_to_dict())
    else:
        taxdump = Taxdump(path_to_taxdump, **data)
    return taxdump


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#!/usr/bin/env python3

"""
BlobToolKit - assembly exploration, QC and filtering.

usage: btk [<command>] [<subcommand>] [<args>...] [-h|--help] [--version]

commands:
    blobtools       BTK command line component
    pipeline        BTK analysis pipeline
    -h, --help      show this
    -v, --version   show version number
See 'btk <command> --help' for more information on a specific command.

"""


import os
import sys

from docopt import DocoptExit
from docopt import docopt
from pkg_resources import working_set
from tolkein import tolog

from .lib.version import __version__

LOGGER = tolog.logger(__name__)


def cli():
    """Entry point."""
    if len(sys.argv) > 2:
        try:
            args = docopt(__doc__, help=False, version=__version__)
        except DocoptExit:
            args = {
                "<command>": " ".join(sys.argv[1:2]),
            }
    elif len(sys.argv) > 1:
        try:
            args = docopt(__doc__, help=False, version=__version__)
        except DocoptExit:
            args = {"<command>": sys.argv[1]}
    else:
        print(__doc__)
    if args["<command>"]:
        args.update({"<tool>": os.path.basename(sys.argv[0])})
        # load <command> from entry_points
        for entry_point in working_set.iter_entry_points("%s.subcmd" % args["<tool>"]):
            if entry_point.name == args["<command>"]:
                subcommand = entry_point.load()
                sys.exit(subcommand())
        LOGGER.error(
            "'%s %s' is not a valid command", args["<tool>"], args["<command>"]
        )
        sys.exit(1)
    raise DocoptExit



"""
A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

import io

# import platform
# import re
from os import makedirs
from os import remove
from os.path import abspath
from os.path import dirname
from os.path import isfile
from os.path import join

from setuptools import find_packages
from setuptools import setup

# from setuptools.command.install import install

# system = platform.system()
# # arch, _ = platform.architecture()
# if system == "Linux":
#     blobtoolkit_api = "bin/blobtoolkit-api-linux"
#     blobtoolkit_viewer = "bin/blobtoolkit-viewer-linux"
# if system == "Windows":
#     blobtoolkit_api = "bin/blobtoolkit-api-win.exe"
#     blobtoolkit_viewer = "bin/blobtoolkit-viewer-win.exe"
# if system == "Darwin":
#     blobtoolkit_api = "bin/blobtoolkit-api-macos"
#     blobtoolkit_viewer = "bin/blobtoolkit-viewer-macos"

# from https://stackoverflow.com/questions/45150304/how-to-force-a-python-wheel-to-be-platform-specific-when-building-it # noqa
# try:
#     from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

#     class bdist_wheel(_bdist_wheel):
#         def finalize_options(self):
#             _bdist_wheel.finalize_options(self)
#             # Mark us as not a pure python package (we have platform specific rust code)
#             self.root_is_pure = False

#         def get_tag(self):
#             # this set's us up to build generic wheels.
#             # note: we're only doing this for windows right now (causes packaging issues
#             # with osx)
#             tag = _bdist_wheel.get_tag(self)
#             print(tag)
#             repl = "macosx_10_6_intel.macosx_10_9_intel.macosx_10_9_x86_64"
#             if tag[2] == "macosx_10_9_x86_64":
#                 tag = (tag[0], tag[1], repl)
#                 return tag

#             if not system == "Windows":
#                 return _bdist_wheel.get_tag(self)

#             python, abi, plat = _bdist_wheel.get_tag(self)
#             python, abi = "py3", "none"
#             return python, abi, plat


# except ImportError:
#     print("Error happens")
#     bdist_wheel = None


# class PostInstallCommand(install):
#     """Post-installation for installation mode."""

#     def run(self):
#         source_dir = dirname(abspath(__file__))
#         makedirs(self.install_scripts, exist_ok=True)

#         self.copy_file(source_dir, blobtoolkit_api, "blobtoolkit-api")
#         self.copy_file(source_dir, blobtoolkit_viewer, "blobtoolkit-viewer")

#     def copy_file(self, source_dir, file_name, executable_name):
#         source = join(source_dir, file_name)
#         target = join(self.install_scripts, executable_name)
#         if isfile(target):
#             remove(target)

#         self.move_file(source, target)


def read(*names, **kwargs):
    """Read file."""
    with io.open(
        join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")
    ) as fh:
        return fh.read()


setup(
    name="blobtoolkit",  # Required
    version="3.2.1",
    description="blobtoolkit",  # Optional
    long_description="blobtoolkit",  # Optional
    long_description_content_type="text/markdown",
    # long_description="%s\n%s"
    # % (
    #     re.compile("^.. start-badges.*^.. end-badges", re.M | re.S).sub(
    #         "", read("README.rst")
    #     ),
    #     re.sub(":[a-z]+:`~?(.*?)`", r"``\1``", read("CHANGELOG.rst")),
    # ),
    # long_description_content_type="text/x-rst",  # Optional (see note above)
    url="https://github.com/blobtoolkit/blobtools2",  # Optional
    # This should be your name or the name of the organization which owns the
    # project.
    author="blobtoolkit",  # Optional
    # This should be a valid email address corresponding to the author listed
    # above.
    author_email="blobtoolkit@genomehubs.org",  # Optional
    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        # Pick your license as you wish
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate you support Python 3. These classifiers are *not*
        # checked by 'pip install'. See instead 'python_requires' below.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
    ],
    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a list of additional keywords, separated
    # by commas, to be used to assist searching for the distribution in a
    # larger catalog.
    keywords="bioinformatics",  # Optional
    # When your source code is in a subdirectory under the project root, e.g.
    # `src/`, it is necessary to specify the `package_dir` argument.
    package_dir={"": "src"},  # Optional
    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(where="src"),  # Required
    # Specify which Python versions you support. In contrast to the
    # 'Programming Language' classifiers above, 'pip install' will check this
    # and refuse to install the project if the version does not match. See
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires=">=3.6, <4",
    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "chromedriver-binary-auto==0.1.1",
        "docopt>=0.6.2",
        "fastjsonschema==2.15.3",
        "geckodriver-autoinstaller==0.1.0",
        "psutil==5.9.0",
        "pysam==0.18.0",
        "pyvirtualdisplay==2.2",
        "pyyaml",
        "selenium==4.1.0",
        "tolkein>=0.4.0",
        "tqdm==4.62.3",
        "ujson>=3.0.0",
        # Additional pipeline dependencies
        "defusedxml==0.7.1",
        "requests==2.27.1",
        "snakemake==6.15.3",
    ],  # Optional
    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    extras_require={  # Optional
        "dev": ["pycodestyle>=2.6.0", "pydocstyle>=5.0.2", "pylint>=2.5.3"],
        "test": [
            "coverage>=5.1",
            "coveralls>=2.0.0",
            "mock>=4.0.2",
            "pytest-cov>=2.10.0",
            "pytest-isort>=5",
            "pytest-mock>=3.1.1",
            "pytest>=6.0.0",
        ],
    },
    entry_points={
        "console_scripts": ["blobtools = blobtools:cli", "btk = btk:cli",],
        "blobtools.subcmd": [
            "add = blobtools.lib.add:cli",
            "create = blobtools.lib.add:cli",
            "filter = blobtools.lib.filter:cli",
            "host = blobtools.lib.host:cli",
            "remove = blobtools.lib.remove:cli",
            "replace = blobtools.lib.add:cli",
            "validate = blobtools.lib.validate:cli",
            "view = blobtools.lib.view:cli",
        ],
        "btk.subcmd": [
            "pipeline = btk_pipeline:main",
            # "blobtools = blobtools:cli",
        ],
        "pipeline.subcmd": [
            "data = btk_pipeline.lib.data:main",
            "run = btk_pipeline.lib.run:main",
            "add-summary-to-metadata = btk_pipeline.lib.add_summary_to_metadata:main",
            "chunk-fasta = btk_pipeline.lib.chunk_fasta:main",
            "count-busco-genes = btk_pipeline.lib.count_busco_genes:main",
            "extract-busco-genes = btk_pipeline.lib.extract_busco_genes:main",
            "generate-config = btk_pipeline.lib.generate_config:main",
            "generate-static-images = btk_pipeline.lib.generate_static_images:main",
            "resume-pipeline = btk_pipeline.lib.resume_pipeline:main",
            "transfer-completed = btk_pipeline.lib.transfer_completed:main",
            "unchunk-blast = btk_pipeline.lib.unchunk_blast:main",
            "window-stats = btk_pipeline.lib.window_stats:main",
        ],
    },
    project_urls={
        "Bug Reports": "https://github.com/blobtoolkit/blobtools2/issues",
        "Source": "https://github.com/blobtoolkit/blobtools2",
    },
    include_package_data=True,
    zip_safe=False,
)



"""This is a template package to demonstrate Continuous Integration."""

from .field import Array, Category, Field, Identifier, MultiArray, Variable
from .taxdump import Taxdump
from .dataset import Metadata



#!/usr/bin/env python3

# pylint: disable=no-member, too-many-branches, too-many-statements, too-many-locals, W0603, W0703

"""
Validate a BlobDir.

Usage:
    blobtools validate [--basic] [--example] DIRECTORY

Arguments:
    DIRECTORY             BlobDir directory.

Options:
    --basic    Only require basic metadata.
    --example  Validate example dataset.
"""

import copy
import os
import os.path
import re
import sys
import urllib

import fastjsonschema
import ujson
from docopt import docopt

from .version import __version__

PIDS = []


def load_json_file(file, schema=False):
    """Load a JSON file."""
    result = []
    error_offset = 7 if schema else 0
    try:
        with open(file, "r") as fh:
            data = ujson.load(fh)
            if "properties" in data:
                for property, value in data["properties"].items():
                    if "$ref" in value:
                        ref_file = os.path.join(os.path.dirname(file), value["$ref"])
                        if os.path.exists(ref_file):
                            value["$ref"] = "file://%s" % ref_file
            return data
    except IOError as err:
        show_error(err.strerror, 1 + error_offset)
    except ValueError as err:
        show_error(err, 2 + error_offset)


def url_scheme(url, path):
    """Treat local URLs as 'file://'."""
    if not urllib.parse(url).scheme:
        url = "file://" + os.path.join(path, url)
    return url


def show_error(error, step=0):
    """List errors then quit."""
    stages = [
        "during undefined stage",
        "while trying to read file",
        "while trying to parse JSON",
        "while testing raw metadata against schema",
        "while testing processed metadata against schema",
        "while generating data schema",
        "while testing data against generated schema",
        "while checking standard fields",
        "while trying to read schema file",
        "while trying to parse JSON schema",
    ]
    if error:
        stage = stages[step]
        print("NOT VALID: Error %s" % stage)
        quit(error)


def flatten_fields(fields, parent={}):
    """Flatten hierarchical field list."""
    field_list = []
    for field in fields:
        for key, value in parent.items():
            if key not in field and key not in ["children", "data"]:
                field[key] = value
        if "children" not in field:
            try:
                check_expected_field_properties(field)
            except AssertionError as err:
                show_error(err, 7)
            field_list.append(field)
        else:
            field_list += flatten_fields(field["children"], field)
        if "data" in field:
            field_list += flatten_fields(field["data"], field)
    return field_list


def check_expected_field_properties(field):
    """Test properties in expected fields."""
    field_id = field["id"]
    if field_id in ["gc", "length", "ncount"] or re.match(r".+_cov$", field_id):
        assert field["type"] == "variable", (
            "expected %s.type to be 'variable'" % field_id
        )
        assert field["range"][0] >= 0, "expected %s.range[0] to be >= 1" % field_id
        if field_id in ["length", "ncount"] or re.match(r".+read_cov$", field_id):
            assert field["datatype"] == "integer", (
                "expected %s.datatype to be 'integer'" % field_id
            )
        else:
            assert field["datatype"] == "float", (
                "expected %s.datatype to be 'float'" % field_id
            )
    elif re.match(r".+_busco", field_id):
        assert field["type"] == "multiarray", (
            "expected %s.type to be 'multiarray'" % field_id
        )
        assert field["datatype"] == "mixed", (
            "expected %s.datatype to be 'mixed'" % field_id
        )
        assert "Busco id" in field["headers"], (
            "expected to find 'Busco id' in %s.headers" % field_id
        )
        assert "Status" in field["headers"], (
            "expected to find 'Status' in %s.headers" % field_id
        )
        assert "category_slot" in field, (
            "expected to find 'category_slot' in %s" % field_id
        )
        message = (
            "expected %s.category_slot to match index of 'Status' in %s.headers"
            % (field_id, field_id)
        )
        assert field["category_slot"] == field["headers"].index("Status"), message
    elif re.match(r"bestsum.+", field_id):
        if re.match(r".+_cindex", field_id):
            assert field["type"] == "variable", (
                "expected %s.type to be 'variable'" % field_id
            )
            assert field["datatype"] == "integer", (
                "expected %s.datatype to be 'integer'" % field_id
            )
            assert field["range"][0] >= 0, "expected %s.range[0] to be >= 1" % field_id
        elif re.match(r".+_positions", field_id):
            assert field["type"] == "multiarray", (
                "expected %s.type to be 'multiarray'" % field_id
            )
            if re.match(r".+_.+_positions", field_id):
                assert field["datatype"] == "string", (
                    "expected %s.datatype to be 'string'" % field_id
                )
                assert "name" in field["headers"], (
                    "expected to find 'name' in %s.headers" % field_id
                )
                assert "category_slot" in field, (
                    "expected to find 'category_slot' in %s" % field_id
                )
                message = (
                    "expected %s.category_slot to match index of 'name' in %s.headers"
                    % (field_id, field_id)
                )
                assert field["category_slot"] == field["headers"].index("name"), message
                assert "linked_field" in field, (
                    "expected to find 'linked_field' in %s" % field_id
                )
            else:
                assert field["datatype"] == "mixed", (
                    "expected %s.datatype to be 'mixed'" % field_id
                )
                assert "taxid" in field["headers"], (
                    "expected to find 'taxid' in %s.headers" % field_id
                )
                assert "score" in field["headers"], (
                    "expected to find 'score' in %s.headers" % field_id
                )
                assert "start" in field["headers"], (
                    "expected to find 'start' in %s.headers" % field_id
                )
                assert "end" in field["headers"], (
                    "expected to find 'end' in %s.headers" % field_id
                )
                assert "subject" in field["headers"], (
                    "expected to find 'subject' in %s.headers" % field_id
                )
                assert "index" in field["headers"], (
                    "expected to find 'index' in %s.headers" % field_id
                )
        elif re.match(r".+_score", field_id):
            assert field["type"] == "variable", (
                "expected %s.type to be 'variable'" % field_id
            )
            assert field["datatype"] == "float", (
                "expected %s.datatype to be 'float'" % field_id
            )
            assert field["range"][0] >= 0, "expected %s.range[0] to be >= 1" % field_id
        else:
            assert field["type"] == "category", (
                "expected %s.type to be 'category'" % field_id
            )


def generate_data_schema(field, meta, data_schemas, data):
    """Modify default schema to match field properties."""
    data_schema = copy.deepcopy(data_schemas[field["type"]])
    schema_values = data_schema["properties"]["values"]
    schema_values["minItems"] = meta["records"]
    schema_values["maxItems"] = meta["records"]
    if field["type"] == "variable":
        schema_values["items"]["minimum"] = field["range"][0]
        schema_values["items"]["maximum"] = field["range"][1]
        value_type = "integer" if field["datatype"] == "integer" else "number"
        schema_values["items"]["type"] = value_type
    elif field["type"] == "category":
        try:
            schema_values["items"]["minimum"] = 0
            schema_values["items"]["maximum"] = len(data["keys"]) - 1
        except KeyError as err:
            show_error(err, 5)
    elif field["type"] == "multiarray":
        schema_values["items"]["items"]["minItems"] = len(field["headers"])
        schema_values["items"]["items"]["maxItems"] = len(field["headers"])
        if "category_slot" in field:
            try:
                schema_values["items"]["items"]["items"] = [
                    {} for x in field["headers"]
                ]
                category = {
                    "type": ["integer", "null"],
                    "minimum": 0,
                    "maximum": len(data["keys"]) - 1,
                }
                schema_values["items"]["items"]["items"][
                    field["category_slot"]
                ] = category
            except KeyError as err:
                show_error(err, 5)
    elif field["type"] == "array":
        schema_values["items"]["minItems"] = len(field["headers"])
        schema_values["items"]["maxItems"] = len(field["headers"])
        if "category_slot" in field:
            try:
                schema_values["items"]["items"] = [{} for x in field["headers"]]
                category = {
                    "type": ["integer", "null"],
                    "minimum": 0,
                    "maximum": len(data["keys"]) - 1,
                }
                schema_values["items"]["items"][field["category_slot"]] = category
            except KeyError as err:
                show_error(err, 5)
    return data_schema


def main(args):
    """Entry point for validator."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    schema_level = "basic" if args["--basic"] else "meta"
    schema = os.path.join(script_dir, "schema", "%s.schema.json" % schema_level)
    file = args["DIRECTORY"]
    if os.path.isdir(file):
        file = os.path.join(file, "meta.json")
    if not os.path.isfile(file):
        if args["--example"]:
            print("Validating example file.")
            file = os.path.join(script_dir, "example", "FXWY01", "meta.json")
        else:
            show_error("%s not found" % file, 1)
    validate = fastjsonschema.compile(load_json_file(schema, True))

    meta = load_json_file(file)

    try:
        validate(meta)
    except fastjsonschema.exceptions.JsonSchemaException as err:
        show_error(err, 3)

    flattened_fields = flatten_fields(meta["fields"])
    try:
        meta["fields"] = flattened_fields
        validate(meta)
    except fastjsonschema.exceptions.JsonSchemaException as err:
        show_error(err, 4)

    validators = {}
    data_schemas = {}
    for type in ["array", "category", "identifier", "multiarray", "variable"]:
        type_schema = os.path.join(
            script_dir, "schema", "subschemas", "%s.meta.schema.json" % type
        )
        validators[type] = fastjsonschema.compile(load_json_file(type_schema, True))
        data_schemas[type] = load_json_file(
            os.path.join(
                script_dir, "schema", "subschemas", "%s.data.schema.json" % type
            ),
            True,
        )

    for field in meta["fields"]:
        try:
            validators[field["type"]](field)
        except fastjsonschema.exceptions.JsonSchemaException as err:
            show_error(err, 4)
        data_file = file.replace("meta.json", "%s.json" % field["id"])
        data = load_json_file(data_file)
        data_schema = generate_data_schema(field, meta, data_schemas, data)
        data_validator = fastjsonschema.compile(data_schema)
        try:
            data_validator(data)
        except fastjsonschema.exceptions.JsonSchemaException as err:
            print(field["id"])
            show_error(err, 6)
    print("VALID")


def cli():
    """Entry point."""
    if len(sys.argv) == sys.argv.index(__name__.split(".")[-1]) + 1:
        args = docopt(__doc__, argv=[])
    else:
        args = docopt(__doc__, version=__version__)
    main(args)


if __name__ == "__main__":
    cli()
